//-------------------------------------------------------------------------------------------------+
// PersonDetection.h
// Define functions of PersonDetection class that detects person by using LRF data and Global map
//-------------------------------------------------------------------------------------------------+

#define _USE_MATH_DEFINES
#include <cmath>
#include "PersonDetection.h"


//------------------------------------------------------------------------------------------------+
// Define of macro
//------------------------------------------------------------------------------------------------+
#define Igrey(i,x,y) ((uchar*)((i)->imageData + (i)->widthStep*(y)))[(x)]		// 24rbgO[XP[


//-------------------------------------------------------------------------------------------------+
// LRF̃f[^ƒn}rĐl炵̂𒊏o
//-------------------------------------------------------------------------------------------------+
TUT::TimedPersonsGlobalPosition PersonDetection::ExtractPerson( SensorRTC::LaserRangeSensor::idl::MeasuredData rangedata, RTC::Pose2D pose ){

	TUT::TimedPersonsGlobalPosition persons;

	if( !image ) return persons;

	vector<floatCoordinate> coordinates;

	IplImage *img = cvCreateImage( cvSize( image->width, image->height ), IPL_DEPTH_8U, 1 );

	cvRectangle( img, cvPoint(0,0), cvPoint(img->width-1,img->height-1), cvScalar(255), CV_FILLED );

	DrowRangeData( img, rangedata, pose );
	coordinates = ImageExOr( img, image, NEIGHBOR8 );
	if( img ) cvReleaseImage( &img );

	persons.data.length( coordinates.size() );

	for( int i=0; i<(int)coordinates.size(); i++ ){	

		persons.data[i].x = ReConvert( coordinates[i].x );
		persons.data[i].y = ReConvert( coordinates[i].y, image->height );
	}

	return persons;
}


//-------------------------------------------------------------------------------------------------+
// Floatn}̕`
//-------------------------------------------------------------------------------------------------+
void PersonDetection::DrowMap( MRFC::TimedFloatAbsoluteOGMapData *mapdata, float threshold, double scale ){

	if( !mapdata || mapdata->cells.length() <= 0 ) return;

	x_scale = mapdata->mapconfig.xScale / scale;
	y_scale = mapdata->mapconfig.yScale / scale;

	origin_x = mapdata->mapconfig.origin.position.x;
	origin_y = mapdata->mapconfig.origin.position.y;

	// 摜𐶐
	IplImage *map_image = cvCreateImage( cvSize( mapdata->mapconfig.width, mapdata->mapconfig.height ), IPL_DEPTH_8U, 1 );

	DrowFloatMap( map_image, mapdata, threshold );

	if( image ) cvReleaseImage( &image );
	image = cvCreateImage( cvSize( scale * mapdata->mapconfig.width, scale * mapdata->mapconfig.height ), IPL_DEPTH_8U, 1 );

	cvResize( map_image, image, CV_INTER_AREA );
	cvReleaseImage( &map_image );
}


//-------------------------------------------------------------------------------------------------+
// Octetn}̕`
//-------------------------------------------------------------------------------------------------+
void PersonDetection::DrowMap( MRFC::TimedAbsoluteOGMapData *mapdata, int threshold, double scale ){

	if( !mapdata || mapdata->cells.length() <= 0 ) return;

	x_scale = mapdata->mapconfig.xScale / scale;
	y_scale = mapdata->mapconfig.yScale / scale;

	origin_x = mapdata->mapconfig.origin.position.x;
	origin_y = mapdata->mapconfig.origin.position.y;

	// 摜𐶐
	IplImage *map_image = cvCreateImage( cvSize( mapdata->mapconfig.width, mapdata->mapconfig.height ), IPL_DEPTH_8U, 1 );

	DrowOctetMap( image, mapdata, threshold );

	if( image ) cvReleaseImage( &image );
	image = cvCreateImage( cvSize( scale * mapdata->mapconfig.width, scale * mapdata->mapconfig.height ), IPL_DEPTH_8U, 1 );

	cvResize( map_image, image, CV_INTER_AREA );
	cvReleaseImage( &map_image );
}


//-------------------------------------------------------------------------------------------------+
// Floatn}̕`
//-------------------------------------------------------------------------------------------------+
void PersonDetection::DrowFloatMap( IplImage *img, MRFC::TimedFloatAbsoluteOGMapData *mapdata, float threshold ){

	for( int i=0; i<img->height; i++ ) for( int x=0; x<img->width; x++){

		int y = img->height - i - 1;

		if( mapdata->cells[x + i * img->width] < 0.0 ) Put( img, x, y, 0 );

		else if( 1.0 < (float)mapdata->cells[x + i * img->width] ) Put( img, x, y, 0 );

		else{
			double value = 1.0 - mapdata->cells[x + i * img->width];

			Put( img, x, y, ( value < threshold )? 0: 255 );
		}
	}
}


//-------------------------------------------------------------------------------------------------+
// Octetn}̕`
//-------------------------------------------------------------------------------------------------+
void PersonDetection::DrowOctetMap( IplImage *img, MRFC::TimedAbsoluteOGMapData *mapdata, int threshold ){

	for( int i=0; i<img->height; i++ ) for( int x=0; x<img->width; x++){

		int y = img->height - i - 1;

		if( mapdata->cells[x + i * img->width] < 0 ) Put( img, x, y, 0 );

		else if( 100 < (int)mapdata->cells[x + i * img->width] ) Put( img, x, y, 0 );

		else{
			int value = 255 - (int)( 2.55 * (double)mapdata->cells[x + i * img->width] );

			Put( img, x, y, ( value < threshold )? 0: 255 );
		}
	}
}


//------------------------------------------------------------------------------------------------+
// f[^̕`
//------------------------------------------------------------------------------------------------+
void PersonDetection::DrowRangeData( IplImage *img, SensorRTC::LaserRangeSensor::idl::MeasuredData rangedata, RTC::Pose2D pose ){

	if( x_scale != y_scale ) return;

	for( int i=0; i<rangedata.distance.length(); i++ ){

		double length = 0.001 * rangedata.distance[i];

		if( length < 0.2 ) continue;

		double theta = rangedata.startPosition -90.0 + (double)i * ( rangedata.endPosition - rangedata.startPosition ) / (double)(rangedata.distance.length()-1);
		theta = theta * M_PI / 180.0 + pose.heading;

		double x = length * cos( theta ) + pose.position.x;
		double y = length * sin( theta ) + pose.position.y;

		Segment( img, x, y, theta, 6.0 );
//		Put( img, Convert( x ), Convert( y, image->height ), 0 );
	}
}


//------------------------------------------------------------------------------------------------+
// wfwFɂ
//------------------------------------------------------------------------------------------------+
void PersonDetection::Put( IplImage *img, int x, int y, int value ){

	if( !img || x < 0 || y < 0 || img->width <= x || img->height <= y ) return;

	Igrey( img, x, y ) = value;
}


//------------------------------------------------------------------------------------------------+
// `
//------------------------------------------------------------------------------------------------+
void PersonDetection::Segment( IplImage *img, double x, double y, double heading, double length, int value ){

	if( !img ) return;

	CvPoint from = { Convert(x), Convert( y, img->height ) };
	CvPoint to   = { from.x + (int)( length * cos( heading )), from.y - (int)( length * sin( heading )) };

	cvLine( img, from, to, cvScalar(value), 1, 8, 0 );
}


//------------------------------------------------------------------------------------------------+
// 2_Ԑ`
//------------------------------------------------------------------------------------------------+
void PersonDetection::ConnectLine( IplImage *img, double x1, double y1, double x2, double y2, int value ){

	if( !img ) return;

	CvPoint from = { Convert(x1), Convert( y1, img->height ) };
	CvPoint to   = { Convert(x2), Convert( y2, img->height ) };

	cvLine( img, from, to, cvScalar(value), 1, 8, 0 );
}


//-------------------------------------------------------------------------------------------------+
// wW珉܂鍕f̘ÄĂ̕ύWԂ
//-------------------------------------------------------------------------------------------------+
floatCoordinate PersonDetection::EraseArea( IplImage *img, int x, int y, Neighbor neighbor, int *size ){

	vector<intCoordinate> check;

	floatCoordinate ave;
	int num = 0;

	check.push_back( intCoordinate( x, y ) );

	while( !check.empty() ){

		x = check.back().x;
		y = check.back().y;
		check.pop_back();

		if( x<0 || y<0 || img->width<=x || img->height<=y || Igrey( img, x, y ) != 0 ) continue;

		Igrey( img, x, y ) = 255;

		ave.x += (float)x;
		ave.y += (float)y;
		num++;

		check.push_back( intCoordinate( x+1, y ));
		check.push_back( intCoordinate( x-1, y ));
		check.push_back( intCoordinate( x, y+1 ));
		check.push_back( intCoordinate( x, y-1 ));

		if( neighbor == NEIGHBOR8 ){

			check.push_back( intCoordinate( x+1, y+1 ));
			check.push_back( intCoordinate( x+1, y-1 ));
			check.push_back( intCoordinate( x-1, y+1 ));
			check.push_back( intCoordinate( x-1, y-1 ));
		}
	}

	if( num != 0 ){

		ave.x /= (float)num;
		ave.y /= (float)num;
	}

	if( size ) *size = num;

	return ave;
}


//-------------------------------------------------------------------------------------------------+
// Q̉摜rďdȂ̖img1̘Ä̍WoĕԂ
//-------------------------------------------------------------------------------------------------+
vector<floatCoordinate> PersonDetection::ImageExOr( IplImage *img1, IplImage *img2, Neighbor neighbor ){

	vector<floatCoordinate> values;

	if( img1->width != img2->width || img1->height != img2->height ) return values;

	// l̉摜c
	IplImage *lrf_img = cvCreateImage( cvGetSize(img2), IPL_DEPTH_8U, 1 );
	cvErode( img1, lrf_img, 0, 4 );

	// 摜𑖍ėsNZTď
	for( int y=0; y<img1->height; y++ ) for( int x=0; x<img1->width; x++ ){

		if( Igrey( lrf_img, x, y ) != 0 || Igrey( img2, x, y ) != 0 ) continue;

		int size=0;

		EraseArea( lrf_img, x, y, neighbor, &size );
	}

	// Äo
	for( int y=0; y<img1->height; y++ ) for( int x=0; x<img1->width; x++ ){

		if( Igrey( lrf_img, x, y ) == 0 ){

			int size=0;
			floatCoordinate data = EraseArea( lrf_img, x, y, neighbor, &size );
			if( size < 550 ) values.push_back( data );
		}
	}

	cvReleaseImage( &lrf_img );

	return values;
}


//------------------------------------------------------------------------------------------------+
// W摜Wɕϊ
//------------------------------------------------------------------------------------------------+
int PersonDetection::Convert( double x ){

	return (int)(( x - origin_x ) / x_scale );
}


//------------------------------------------------------------------------------------------------+
// W摜Wɕϊ
//------------------------------------------------------------------------------------------------+
int PersonDetection::Convert( double y, int height ){

	return ( height - (int)(( y - origin_y ) / y_scale ));
}


//------------------------------------------------------------------------------------------------+
// 摜WWɕϊ
//------------------------------------------------------------------------------------------------+
double PersonDetection::ReConvert( double x ){

	return x * x_scale + origin_x;
}


//------------------------------------------------------------------------------------------------+
// 摜WWɕϊ
//------------------------------------------------------------------------------------------------+
double PersonDetection::ReConvert( double y, int height ){

	return ( (double)height - y ) * y_scale + origin_y;
}
