//-------------------------------------------------------------------------------------------------+
// PersonDetector.cpp
// This RTC detects persons by using Globalmap and a data from LRF. 
//-------------------------------------------------------------------------------------------------+

#include "PersonDetector.h"
#include <iostream>

using namespace std;


//------------------------------------------------------------------------------------------------+
// Module specification
//------------------------------------------------------------------------------------------------+
static const char* PersonDetector_spec[] =
{
    "implementation_id", "PersonDetector",
    "type_name",         "PersonDetector",
    "description",       "This RTC is used to show Global Path.",
    "version",           "1.0.0",
    "vendor",            "aisl.cs.tut.ac.jp",
    "category",          "Viewe",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "0",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.threshold", "0.2",
    "conf.default.map_type", "1",
    // Widget
    "conf.__widget__.threshold", "text",
    "conf.__widget__.map_type", "text",
    ""
};


//-------------------------------------------------------------------------------------------------+
// constructor
//-------------------------------------------------------------------------------------------------+
PersonDetector::PersonDetector(RTC::Manager* manager)
  : RTC::DataFlowComponentBase(manager),
    m_rangedataIn("rangedata", m_rangedata),
    m_robot_poseIn("robot_pose", m_robot_pose),
    m_estimated_poseIn("estimated_pose", m_estimated_pose),
    m_personsOut("persons", m_persons),
    m_AbsoluteMapServicePort("AbsoluteMapService")
{
	cerr << "------------------------------------------------------------+\n"
		 << " PersonDetector Component (Ver. 7th February 2012)\n"
		 << "------------------------------------------------------------+\n\n";

	this->manager = manager;
}


//-------------------------------------------------------------------------------------------------+
// destructor
//-------------------------------------------------------------------------------------------------+
PersonDetector::~PersonDetector(){

	if( convert ) delete convert;
}


//-------------------------------------------------------------------------------------------------+
// vON̏֐
//-------------------------------------------------------------------------------------------------+
RTC::ReturnCode_t PersonDetector::onInitialize(){

	// Set InPort buffers
	addInPort("rangedata", m_rangedataIn);
	addInPort("robot_pose", m_robot_poseIn);
	addInPort("estimated_pose", m_estimated_poseIn);

	// Set OutPort buffers
	addOutPort("persons", m_personsOut);

	// Set service consumers to Ports
	m_AbsoluteMapServicePort.registerConsumer("AbsoluteMapService", "MRFC::AbsoluteMapService", m_AbsoluteMapService);

	// Set CORBA Service Ports
	addPort(m_AbsoluteMapServicePort);

	// Bind variables and configuration variable
	bindParameter("threshold", m_threshold, "0.2");
	bindParameter("map_type", m_map_type, "1");

	return RTC::RTC_OK;
}


//-------------------------------------------------------------------------------------------------+
// ̏
//-------------------------------------------------------------------------------------------------+
RTC::ReturnCode_t PersonDetector::onActivated(RTC::UniqueId ec_id){

	if( convert ) delete convert;
	convert = new PersonDetection;

	input_range = input_robot = input_estimated_pose = false;

	return RTC::RTC_OK;
}


//-------------------------------------------------------------------------------------------------+
// 񊈐̏
//-------------------------------------------------------------------------------------------------+
RTC::ReturnCode_t PersonDetector::onDeactivated(RTC::UniqueId ec_id){

	delete convert;
	convert = 0;

	return RTC::RTC_OK;
}


//-------------------------------------------------------------------------------------------------+
// s֐
//-------------------------------------------------------------------------------------------------+
RTC::ReturnCode_t PersonDetector::onExecute(RTC::UniqueId ec_id){

	if( !CheckConnectServicePort() ){

		Sleep( 1000 );
		return RTC::RTC_OK;
	}

	InputRange();
	InputRobot();

	if( input_range && input_robot ){

		DrowMap();
		input_range = input_robot = false;
	}

	return RTC::RTC_OK;
}


//-------------------------------------------------------------------------------------------------+
// n}LRFf[^摜ɕ`悵Ĕr邱Ƃɂlʒǔo
//-------------------------------------------------------------------------------------------------+
void PersonDetector::DrowMap(){

	RTC::OGMapConfig conf = m_AbsoluteMapService->getAbsoluteOGMapConfig();

	if( m_map_type ){

		MRFC::TimedFloatAbsoluteOGMapData *mapdata = m_AbsoluteMapService->getFloatAbsoluteOGMap( conf.origin.position.x, conf.origin.position.y, conf.width, conf.height );
		convert->DrowMap( mapdata, m_threshold, 2.0 );
		delete mapdata;
	}
	else{
		MRFC::TimedAbsoluteOGMapData *mapdata = m_AbsoluteMapService->getAbsoluteOGMap( conf.origin.position.x, conf.origin.position.y, conf.width, conf.height );
		convert->DrowMap( mapdata, (int)(255.0 * m_threshold), 2.0 );
		delete mapdata;
	}

	if( input_range && input_robot ){

		m_persons = convert->ExtractPerson( m_rangedata.data, m_robot_pose.data );
		m_persons.tm = m_robot_pose.tm;
		m_personsOut.write();
	}
}


//------------------------------------------------------------------------------------------------+
// f[^̓
//------------------------------------------------------------------------------------------------+
void PersonDetector::InputRange(){

	if( m_rangedataIn.isNew() ){

		while( !m_rangedataIn.isEmpty() ) m_rangedataIn.read();
		input_range = true;
	}
}


//------------------------------------------------------------------------------------------------+
// {bg̓
//------------------------------------------------------------------------------------------------+
void PersonDetector::InputRobot(){

	if( m_robot_poseIn.isNew() ){

		while( !m_robot_poseIn.isEmpty() ) m_robot_poseIn.read();
		
		if( !CheckConnectPort(2) ) input_robot = true;

		else if( m_estimated_poseIn.isNew() ){

			while( !m_estimated_poseIn.isEmpty() ) m_estimated_poseIn.read();

			m_robot_pose.data  = GlobalOdometry( m_robot_pose.data );
			m_robot_pose.error = m_estimated_pose.error;

			input_robot = true;
		}
	}
}


//-------------------------------------------------------------------------------------------------+
// IhgƂɌ݈ʒuC
//-------------------------------------------------------------------------------------------------+
Pose2D PersonDetector::GlobalOdometry( Pose2D odometry ){

	double theta = m_estimated_pose.ododata.heading - m_estimated_pose.estdata.heading;

	double dx = odometry.position.x - m_estimated_pose.ododata.position.x;
	double dy = odometry.position.y - m_estimated_pose.ododata.position.y;
	double dt = odometry.heading    - m_estimated_pose.ododata.heading;

	double global_dx =  dx * cos( theta ) + dy * sin( theta );
	double global_dy = -dx * sin( theta ) + dy * cos( theta );

	odometry.position.x = m_estimated_pose.estdata.position.x + global_dx;
	odometry.position.y = m_estimated_pose.estdata.position.y + global_dy;
	odometry.heading    = m_estimated_pose.estdata.heading    + dt;

	return odometry;
}


//-------------------------------------------------------------------------------------------------+
// e|[g̐ڑmF
// ڑĂ΂̐ڑԂ
//-------------------------------------------------------------------------------------------------+
int PersonDetector::CheckConnectPort( int no ){

	RTC::RTObject_var rtobj = RTC::RTObject::_narrow(manager->getPOA()->servant_to_reference(this));

	// Get the port list of the component
	PortServiceList* portlist = rtobj->get_ports();

	if( no < 0 || portlist->length() <= no ) return 0;

	PortService_ptr port = (*portlist)[no];

	return (int)port->get_port_profile()->connector_profiles.length();
}


//-------------------------------------------------------------------------------------------------+
// T[rX|[g̐ڑmF
// ڑĂtrueԂ
//-------------------------------------------------------------------------------------------------+
bool PersonDetector::CheckConnectServicePort(){

	if( CORBA::is_nil(m_AbsoluteMapService.getObject()) ){

		cerr << "Service port is not conected with service provider.\n\n";
		return false;
	}

	if( m_AbsoluteMapService.getObject()->_non_existent() ){
	
		cerr << "Service provider is not activated.\n\n";
		return false;
	}

	return true;
}



//-------------------------------------------------------------------------------------------------+
extern "C"
{
 
  void PersonDetectorInit(RTC::Manager* manager)
  {
    coil::Properties profile(PersonDetector_spec);
    manager->registerFactory(profile,
                             RTC::Create<PersonDetector>,
                             RTC::Delete<PersonDetector>);
  }
};
