//-------------------------------------------------------------------------------------------------+
// PersonDetector.h
// This RTC detects persons by using Globalmap and a data from LRF. 
//-------------------------------------------------------------------------------------------------+

#ifndef PersonDetector_H
#define PersonDetector_H

#define _USE_MATH_DEFINES

#include <rtm/Manager.h>
#include <rtm/DataFlowComponentBase.h>
#include <rtm/CorbaPort.h>
#include <rtm/DataInPort.h>
#include <rtm/DataOutPort.h>
#include <rtm/idl/BasicDataTypeSkel.h>
#include <rtm/idl/ExtendedDataTypesSkel.h>
#include <rtm/idl/InterfaceDataTypesSkel.h>

// Service Consumer stub headers
#include "LaserRangeSensorDataStub.h"
#include "MRFCStub.h"
#include "IIS2Stub.h"
#include "AislGlobalPlanningStub.h"

#include "cv.h"
#include "highgui.h"

#include "PersonDetection.h"

using namespace RTC;
using namespace IIS;
using namespace MRFC;


//-------------------------------------------------------------------------------------------------+
// PersonDetector
//-------------------------------------------------------------------------------------------------+
class PersonDetector : public RTC::DataFlowComponentBase{
public:
	PersonDetector(RTC::Manager* manager);
	~PersonDetector();
	virtual RTC::ReturnCode_t onInitialize();
	virtual RTC::ReturnCode_t onActivated(RTC::UniqueId ec_id);
	virtual RTC::ReturnCode_t onDeactivated(RTC::UniqueId ec_id);
	virtual RTC::ReturnCode_t onExecute(RTC::UniqueId ec_id);

protected:
	double m_threshold;
	int m_map_type;

	// DataInPort declaration
	SensorRTC::LaserRangeSensor::idl::TimedMeasuredData m_rangedata;
	InPort<SensorRTC::LaserRangeSensor::idl::TimedMeasuredData> m_rangedataIn;
	IIS::TimedPose2D m_robot_pose;
	InPort<IIS::TimedPose2D> m_robot_poseIn;
	TimedEstimatedPose2D m_estimated_pose;
	InPort<TimedEstimatedPose2D> m_estimated_poseIn;

	// DataOutPort declaration
	TUT::TimedPersonsGlobalPosition m_persons;
	OutPort<TUT::TimedPersonsGlobalPosition> m_personsOut;

	RTC::CorbaPort m_AbsoluteMapServicePort;
	RTC::CorbaConsumer<MRFC::AbsoluteMapService> m_AbsoluteMapService;


private:
	bool input_range, input_robot, input_estimated_pose;
	PersonDetection *convert;
	RTC::Manager* manager;

	void DrowMap();
	void InputRange();
	void InputRobot();
	Pose2D GlobalOdometry( Pose2D odometry );
	int CheckConnectPort( int no );
	bool CheckConnectServicePort();
};



//-------------------------------------------------------------------------------------------------+
extern "C"
{
  DLL_EXPORT void PersonDetectorInit(RTC::Manager* manager);
};

#endif // PersonDetector_H
